import gc
import subprocess
import sys
import weakref
from unittest.mock import MagicMock, patch, call

from pyx3270.offline import PyX3270Manager


@patch('subprocess.Popen')
def test_pyx3270_manager_init_and_terminate(mock_popen, x3270_cmd_instance):
    mock_process = MagicMock()
    mock_popen.return_value = mock_process
    mock_process.poll.return_value = None  # Processo em execução

    manager = PyX3270Manager(x3270_cmd_instance)

    mock_popen.assert_called_once_with(
        [
            'pyx3270',
            'replay',
            '--directory',
            './screens',
            '--no-tls',
            '--no-emulator',
        ],
        stdin=subprocess.PIPE,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        text=True,
        bufsize=0,
    )
    assert manager.process == mock_process

    manager.terminate()
    mock_process.poll.assert_called_once()
    mock_process.terminate.assert_called_once()
    mock_process.wait.assert_called_once_with(timeout=5)


@patch('subprocess.Popen')
@patch('pyx3270.offline.logger')
def test_pyx3270_manager_exec(mock_logger, mock_popen, x3270_cmd_instance):
    mock_process = MagicMock()
    mock_popen.return_value = mock_process
    mock_process.poll.return_value = None  # Processo em execução

    manager = PyX3270Manager(x3270_cmd_instance)
    manager.emu = MagicMock()  # Mock atributos de emu

    command = 'test_command'
    manager._exec(command)

    mock_logger.info.assert_called_with(
        f'[+] Enviando comando offline: {command}'
    )
    mock_process.stdin.write.assert_called_once_with(f'{command}\n')
    mock_process.stdin.flush.assert_called_once()
    assert manager.emu.pf.call_count == 2
    manager.emu.pf.assert_has_calls([call(1), call(1)])


@patch('subprocess.Popen')
@patch('pyx3270.offline.logger')
def test_pyx3270_manager_exec_inactive_process(
    mock_logger, mock_popen, x3270_cmd_instance
):
    mock_process = MagicMock()
    mock_popen.return_value = mock_process
    mock_process.poll.return_value = 1  # Process is inactive

    manager = PyX3270Manager(x3270_cmd_instance)
    manager.emu = MagicMock()

    command = 'test_command'
    manager._exec(command)

    mock_logger.info.assert_called_with(
        '[!] O processo inativo, não é possível enviar comandos.'
    )
    mock_process.stdin.write.assert_not_called()


def test_pyx3270_manager_next(x3270_cmd_instance):
    manager = PyX3270Manager(x3270_cmd_instance)
    manager._exec = MagicMock()
    manager.next()
    manager._exec.assert_called_once_with('next')


def test_pyx3270_manager_prev(x3270_cmd_instance):
    manager = PyX3270Manager(x3270_cmd_instance)
    manager._exec = MagicMock()
    manager.prev()
    manager._exec.assert_called_once_with('prev')


def test_pyx3270_manager_send_pf(x3270_cmd_instance):
    manager = PyX3270Manager(x3270_cmd_instance)
    manager.next = MagicMock()
    manager.prev = MagicMock()

    manager.send_pf(4)
    manager.next.assert_called_once()
    manager.next.reset_mock()

    manager.send_pf(8)
    manager.next.assert_called_once()
    manager.next.reset_mock()

    manager.send_pf(3)
    manager.prev.assert_called_once()
    manager.prev.reset_mock()

    manager.send_pf(7)
    manager.prev.assert_called_once()
    manager.prev.reset_mock()

    manager.send_pf(1)
    manager.next.assert_not_called()
    manager.prev.assert_not_called()


def test_pyx3270_manager_set_screen(x3270_cmd_instance):
    manager = PyX3270Manager(x3270_cmd_instance)
    manager._exec = MagicMock()
    screen_name = 'test_screen'
    result = manager.set_screen(screen_name)
    manager._exec.assert_called_once_with(f'set {screen_name}')
    assert result is True


def test_pyx3270_manager_change_directory(x3270_cmd_instance):
    manager = PyX3270Manager(x3270_cmd_instance)
    manager._exec = MagicMock()
    manager.emu = MagicMock()
    directory_name = 'new_directory'
    manager.change_directory(directory_name)
    manager._exec.assert_called_once_with(f'change directory {directory_name}')
    manager.emu.pf.assert_called_once_with(1)


def test_pyX3270_manager_del(x3270_cmd_instance):
    manager = PyX3270Manager(x3270_cmd_instance)
    terminate_mock = MagicMock()
    manager.terminate = terminate_mock

    called_flag = {'called': False}

    def on_finalize():
        called_flag['called'] = terminate_mock.called

    weakref.finalize(manager, on_finalize)

    del manager
    gc.collect()

    assert called_flag['called']


@patch('subprocess.Popen')
def test_terminate_kills_process_on_timeout(mock_popen, x3270_cmd_instance):
    # Mock do processo
    mock_process = MagicMock()
    mock_popen.return_value = mock_process

    # Simula que o processo está rodando
    mock_process.poll.return_value = None

    # Simula o wait lançando TimeoutExpired
    mock_process.wait.side_effect = subprocess.TimeoutExpired(
        cmd='testcmd', timeout=5
    )

    # Instancia o manager
    manager = PyX3270Manager(x3270_cmd_instance)

    # Mock do emu.terminate
    manager.emu = MagicMock()

    # Chama o método terminate
    manager.terminate()

    # Verificações
    mock_process.poll.assert_called_once()
    manager.emu.terminate.assert_called_once()
    mock_process.terminate.assert_called_once()
    mock_process.wait.assert_called_once_with(timeout=5)
    mock_process.kill.assert_called_once()



import os
import socket
import sys
import threading
from time import sleep
from typing import Callable

import rich
import typer

from pyx3270 import state
from pyx3270.emulator import X3270
from pyx3270.server import (
    load_screens,
    record_handler,
    replay_handler,
    server_stop,
    start_command_process,
)

app = typer.Typer()


def start_sock(port: int) -> socket.socket:
    tnsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tnsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if os.name == 'posix':
        tnsock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    tnsock.bind(('', port))
    tnsock.listen(5)
    return tnsock


def start_server_thread(
    port: int,
    handler: Callable,
    handler_args: tuple | list | None = None,
    label: str = 'Servidor',
) -> threading.Thread:
    def server_loop():
        tnsock = start_sock(port)
        rich.print(f'[+] {label} escutando na porta {port}')
        while True:
            clientsock, addr = tnsock.accept()
            rich.print(f'[+] Cliente conectado: {addr}')

            th = threading.Thread(
                target=handler, args=(clientsock, *handler_args), daemon=True
            )
            th.start()

    th = threading.Thread(target=server_loop, daemon=True)
    th.start()
    return th


def control_replay(th: threading.Thread) -> None:
    # Aguarda encerramento da thread ou falha do servidor
    while th.is_alive():
        if server_stop.is_set():
            rich.print('[x] Conexão encerrada.')
            if state.command_process:
                state.command_process.terminate()
            server_stop.clear()
            break
        sleep(1)

    while True:
        rich.print('[?] Digite "Q" para sair ou "S" para continuar: ', end='')
        op = input().strip().upper()
        if op == 'Q':
            rich.print('[*] Encerrando aplicação...')
            os._exit(0)
        elif op != 'S':
            rich.print('[!] Opção inválida. Continuando...')
        else:
            rich.print('[*] Reiniciando...')
            break
        sleep(1)

    rich.print('[+] Escutando localhost')


@app.command()
def replay(
    directory: str = typer.Option(default='./screens'),
    port: int = typer.Option(default=3270),
    tls: bool = typer.Option(default=False),
    model: str = typer.Option(default='2'),
    emulator: bool = typer.Option(default=True),
):
    screens = load_screens(directory)
    rich.print(f'[+] REPLAY do caminho: {directory}')

    try:
        while True:
            server_thread = start_server_thread(
                port,
                replay_handler,
                handler_args=(screens, emulator, directory),
                label='Servidor de replay',
            )

            if emulator:
                emu = X3270(visible=True, model=model, save_log_file=True)
                emu.connect_host('localhost', port, tls, mode_3270=False)
                sleep(2)

            start_command_process()
            control_replay(server_thread)
    except KeyboardInterrupt:
        rich.print('\n[x] Interrompido pelo usuário.')
        state.command_process.terminate()
        if emulator:
            emu.terminate()
        sys.exit(0)


@app.command()
def record(
    address: str = typer.Option(),
    directory: str = typer.Option(default='./screens'),
    tls: bool = typer.Option(default=True),
    model: str = typer.Option(default='2'),
    emulator: bool = typer.Option(default=True),
):
    host, *port = address.split(':', 2)
    port = int(*port) if port else 3270

    rich.print(f'[+] RECORD na porta {port}')
    reconnect = False
    try:
        while True:
            if emulator and not reconnect:
                emu = X3270(visible=True, model=model, save_log_file=True)
            elif not reconnect:
                emu = None

            server_thread = start_server_thread(
                port=port,
                handler=record_handler,
                handler_args=(emu, address, directory, 0.01),
                label='Servidor de gravação',
            )

            if emulator:
                rich.print('[+] Conectando ao emulador...')
                if reconnect:
                    emu.reconnect_host()
                else:
                    emu.connect_host('localhost', port, tls, mode_3270=False)

            rich.print(f'[+] Escutando localhost, origem {host=} {port=}')
            control_replay(server_thread)
            reconnect = True
    except KeyboardInterrupt:
        rich.print('\n[x] Interrompido pelo usuário.')
        if emulator:
            emu.terminate()
        sys.exit(0)


if __name__ == '__main__':
    app()



import multiprocessing
import os
import queue
import re
import select
import socket
import sys
import threading
from dataclasses import dataclass
from logging import getLogger

import rich

from pyx3270 import state, tn3270
from pyx3270.emulator import BINARY_FOLDER, X3270
from pyx3270.exceptions import NotConnectedException

logger = getLogger(__name__)
server_stop = threading.Event()
command_queue = multiprocessing.Queue()


@dataclass
class ReplayState:
    screens: dict
    screens_list: list
    current_screen: int
    clear: bool


def ensure_dir(path: str | None) -> None:
    if path and not os.path.isdir(path):
        logger.info(f'[+] Criando diretório: {path}')
        os.makedirs(path)


def is_screen_tn3270(data: bytes) -> bool:
    min_file_len = 100
    return len(data) > min_file_len and b'\x11' in data


def load_screens_basic(record_dir: str) -> dict:
    """Carrega todos os arquivos .bin do diretório e subdiretórios."""
    ensure_dir(record_dir)  # Garante que o diretório existe
    screens = {}

    for root, _, files in os.walk(record_dir):  # Percorre pastas e subpastas
        for f in sorted(files):
            if f.endswith('.bin'):
                file_path = os.path.join(root, f)
                with open(file_path, 'rb') as fd:
                    data = fd.read()
                    # Garante que termina com tn3270.IAC TN_EOR
                    if not data.endswith(tn3270.IAC + tn3270.TN_EOR):
                        data += tn3270.IAC + tn3270.TN_EOR
                    screens[str(f).upper().replace('.BIN', '')] = data

    return screens


def load_screens(record_dir: str) -> dict:
    """Carrega arquivos .bin mantendo a ordem dentro de cada subdiretório."""
    ensure_dir(record_dir)  # Garante que o diretório existe

    screens = {}

    try:
        has_bin = any(f.endswith('.bin') for f in os.listdir(record_dir))
        if not has_bin:
            return load_screens_basic(BINARY_FOLDER)

        for root, _, files in os.walk(record_dir):
            for f in sorted(files):
                if not f.endswith('.bin'):
                    continue

                file_path = os.path.join(root, f)
                with open(file_path, 'rb') as fd:
                    data = fd.read()

                # Garante que termina com tn3270.IAC + TN_EOR
                if not data.endswith(tn3270.IAC + tn3270.TN_EOR):
                    data += tn3270.IAC + tn3270.TN_EOR

                key = f.upper().replace('.BIN', '')
                screens[key] = data

    except Exception:
        logger.error(f'Falha ao carregar caminho: {record_dir}')

    return screens


def find_directory(base_dir: str, search_name: str) -> str | None:
    """
    Busca um diretório dentro da pasta base, considerando
    maiúsculas/minúsculas e busca parcial.
    """
    search_name = (
        search_name.lower()
    )  # Normaliza para comparação case-insensitive

    for dir_name in os.listdir(base_dir):
        if os.path.isdir(os.path.join(base_dir, dir_name)):
            if (
                search_name in dir_name.lower()
            ):  # Compara de forma insensível a maiúsculas/minúsculas
                return os.path.join(base_dir, dir_name)

    return None


def convert_s(hex_string: str) -> str:
    pattern = r'SF\((.*?)\)'

    def replace_match(match):
        parameters = match.group(1).split(',')
        converted_values = [
            f'1D{param.split("=")[1]}' for param in parameters if '=' in param
        ]
        return ''.join(converted_values)

    converted_string = re.sub(pattern, replace_match, hex_string)
    converted_string = converted_string.replace(' ', '')
    return converted_string


def connect_serversock(
    clientsock: socket.socket, address: str
) -> socket.socket:
    host, *port = address.split(':', 2)
    port = int(*port) if port else 3270
    try:
        serversock = socket.create_connection((host, port), timeout=5)
    except Exception as e:
        logger.error(f'[!] Proxy -> MF Falha de conexão: {e}')
        clientsock.close()
        return
    return serversock


def record_data(data_block, record_dir: str, counter: int) -> int:
    if not is_screen_tn3270(data_block):
        logger.warning('[!] Dados recebidos não são uma tela TN3270.')
        return 0
    fn = os.path.join(record_dir, f'{counter:03}.bin')
    with open(fn, 'wb') as f:
        logger.info(f'[+] Gravação de tela: {fn}')
        f.write(data_block)
    return 1


def record_handler(
    clientsock: socket.socket,
    emu: X3270,
    address: str,
    record_dir=None,
    delay=0.01,
) -> None:
    logger.info(f'Iniciando gravação para {clientsock.getpeername()}')
    serversock = connect_serversock(clientsock, address)

    if not serversock:
        logger.error('[!] Falha ao conectar ao servidor.')
        return

    ensure_dir(record_dir)

    counter = 0

    socks = [clientsock, serversock]
    channel = {clientsock: serversock, serversock: clientsock}
    buffers = {clientsock: b'', serversock: b''}
    screens = []

    try:
        while True:
            ready_socks, _, _ = select.select(socks, [], [], delay)
            for s in ready_socks:
                save = True
                data = s.recv(2048)
                if not data:
                    raise ConnectionResetError

                if emu.tls:
                    channel[s].sendall(data)
                    continue

                buffers[s] += data

                if s == serversock and record_dir:
                    while tn3270.IAC + tn3270.TN_EOR in buffers[s]:
                        block, _, rest = buffers[s].partition(
                            tn3270.IAC + tn3270.TN_EOR
                        )
                        full_block = block + tn3270.IAC + tn3270.TN_EOR
                        counter += record_data(full_block, record_dir, counter)
                        buffers[s] = rest
                channel[s].sendall(data)

            if emu.tls:
                buffer = emu.readbuffer('ebcdic')
                if (
                    not buffer.replace(' ', '').replace('0', '')
                    or not save
                    or buffer in screens
                ):
                    continue
                buffer_hex = convert_s(buffer)
                buffer_hex = bytes.fromhex(buffer_hex)
                full_block = (
                    tn3270.START_SCREEN
                    + buffer_hex
                    + tn3270.IAC
                    + tn3270.TN_EOR
                )
                save = False
                screens.append(buffer)
                counter += record_data(full_block, record_dir, counter)

    except (ConnectionResetError, OSError, NotConnectedException):
        logger.info('[!] Conexão encerrada pelo servidor ou erro de rede.')
        for sock in socks:
            sock.close()
        server_stop.set()


def backend_3270(
    clientsock: socket.socket,
    screens: list[bytes],
    current_screen: int,
    emulator: bool,
) -> dict[str, int | bool]:
    aid = None

    while aid not in tn3270.AIDS:
        try:
            aid = clientsock.recv(1)
            if not aid:
                raise ConnectionResetError('Terminal fechado.')
        except socket.timeout:
            continue

    press = clientsock.recv(1)

    # Verifica se press contém um código esperado
    key_press = press and press not in tn3270.AIDS
    clear = False

    if aid in {tn3270.PF3, tn3270.PF7} and key_press and emulator:
        current_screen = max(0, current_screen - 1)
        logger.info('[!] Comando de retorno recebido.')
    elif (
        aid in {tn3270.PF4, tn3270.PF8, tn3270.ENTER}
        and key_press
        and emulator
    ):
        logger.info('[!] Comando de paginação/confirmação recebido.')
        current_screen = min(len(screens) - 1, current_screen + 1)
    elif aid == tn3270.CLEAR and key_press and emulator or aid == tn3270.PF12:
        logger.info('[!] Comando CLEAR recebido.')
        clientsock.sendall(tn3270.CLEAR_SCREEN_BUFFER)
        clear = True

    return dict(current_screen=current_screen, clear=clear)


def listen_for_commands(command_queue):
    logger.info('[*] Aguardando comandos...')
    sys.stdin = open(0, mode='r', encoding='utf-8')
    try:
        while True:
            command = input('[?] Digite um comando: ').strip().lower()
            if command == 'q':
                rich.print("[*] Comando 'quit' recebido, encerrando...")
                break
            command_queue.put(command)
    except (OSError, EOFError) as e:
        logger.warning(f'[*] stdin para comandos fechado: {e}')
    finally:
        logger.warning('[*] Encerrando processo de escuta.')


def handle_set(command: str, screens: dict) -> int | None:
    screen_name = command.split(' ', 1)[1].upper()
    for i, key in enumerate(screens.keys()):
        if screen_name in key:
            rich.print(f'[+] Mudando para a tela: {screen_name}')
            return i
    return None


def handle_add(command: str, screens: dict, screens_list: list) -> None:
    try:
        screen_name, screen_data = command.split(' ', 2)[1:]
        screen_name = screen_name.upper()
        screen_bytes = bytes.fromhex(screen_data)
        final_bytes = (
            tn3270.START_SCREEN + screen_bytes + tn3270.IAC + tn3270.TN_EOR
        )
        screens[screen_name] = final_bytes
        screens_list.append(final_bytes)
    except ValueError:
        logger.warning('[!] Formato inválido ao adicionar tela')


def handle_change_directory(
    command: str, base_directory: str
) -> tuple[dict, list]:
    new_dir = find_directory(base_directory, command.split(' ', 2)[2].strip())
    if new_dir and os.path.isdir(new_dir):
        new_screens = load_screens_basic(new_dir)
        logger.info(f'[+] Mudando para o diretório de telas: {new_dir}')
        return new_screens, list(new_screens.values())
    else:
        logger.info(f'[!] Diretório inválido: {new_dir}')
        return dict(), list()


def process_command(
    command: str,
    clientsock: socket.socket,
    base_directory: str,
    state: ReplayState,
) -> ReplayState:
    command_log = (
        command
        if not command.startswith('add ')
        else ' '.join(command.split()[:2])
    )
    logger.info(f'Comando recebido: {command_log}')

    if command.startswith('set '):
        index = handle_set(command, state.screens)
        if index is not None:
            state.current_screen = index

    elif command.startswith('add '):
        handle_add(command, state.screens, state.screens_list)

    elif command.startswith('change directory '):
        state.screens, state.screens_list = handle_change_directory(
            command, base_directory
        )
        state.current_screen = 0

    elif command == 'next':
        state.current_screen = min(
            len(state.screens), state.current_screen + 1
        )
        logger.info(f"[!] Comando 'next' enviado: {state.current_screen}")

    elif command == 'prev':
        state.current_screen = max(0, state.current_screen - 1)
        logger.info(f"[!] Comando 'prev' enviado: {state.current_screen}")

    elif command == 'clear':
        state.clear = True
        clientsock.sendall(tn3270.CLEAR_SCREEN_BUFFER)
        logger.info(f"[!] Comando 'clear' enviado: {state.current_screen}")

    elif command == 'q':
        logger.info("[!] Comando 'quit' enviado.")
        clientsock.close()

    return state


def start_command_process():
    if state.command_process and state.command_process.is_alive():
        logger.warning('[!] Processo de comando já está em execução.')
        return

    if not multiprocessing.get_start_method(allow_none=True):
        if os.name == 'posix':
            multiprocessing.set_start_method('fork')
        else:
            multiprocessing.set_start_method('spawn')

    logger.info('[+] Iniciando processo de comando...')
    state.command_process = multiprocessing.Process(
        target=listen_for_commands, args=(command_queue,)
    )
    state.command_process.start()


def replay_handler(
    clientsock: socket.socket,
    screens: dict,
    emulator: bool,
    base_directory: str,
) -> None:
    state = ReplayState(
        screens=screens,
        screens_list=list(screens.values()),
        current_screen=0,
        clear=False,
    )

    try:
        peer_name = clientsock.getpeername()
        logger.info(f'Iniciando replay para {peer_name}')
        clientsock.sendall(b'\xff\xfd\x18\xff\xfb\x18')

        while True:
            try:
                if not state.clear:
                    clientsock.sendall(
                        state.screens_list[state.current_screen]
                    )
            except IndexError:
                state.current_screen = len(state.screens_list) - 1
                clientsock.sendall(state.screens_list[state.current_screen])

            try:
                command = command_queue.get_nowait()
                state = process_command(
                    command, clientsock, base_directory, state
                )
            except queue.Empty:
                pass

            result = backend_3270(
                clientsock, state.screens_list, state.current_screen, emulator
            )

            state.current_screen = result.get(
                'current_screen', state.current_screen
            )
            state.clear = result.get('clear', state.clear)

    except Exception as e:
        logger.error(f'[!] Erro fora do esperado: {str(e)}')
        server_stop.set()
    finally:
        clientsock.close()
        server_stop.set()



import pytest

from pyx3270.exceptions import (
    BrokenPipeError,
    CommandError,
    FieldTruncateError,
    KeyboardStateError,
    NotConnectedException,
    TerminatedError,
)


def test_broken_pipe_error():
    with pytest.raises(BrokenPipeError, match='Test BrokenPipeError'):
        raise BrokenPipeError('Test BrokenPipeError')


def test_command_error():
    with pytest.raises(CommandError, match='Test CommandError'):
        raise CommandError('Test CommandError')


def test_terminated_error():
    with pytest.raises(TerminatedError, match='Test TerminatedError'):
        raise TerminatedError('Test TerminatedError')


def test_keyboard_state_error():
    with pytest.raises(KeyboardStateError, match='Test KeyboardStateError'):
        raise KeyboardStateError('Test KeyboardStateError')


def test_field_truncate_error():
    with pytest.raises(FieldTruncateError, match='Test FieldTruncateError'):
        raise FieldTruncateError('Test FieldTruncateError')


def test_not_connected_exception():
    with pytest.raises(
        NotConnectedException, match='Test NotConnectedException'
    ):
        raise NotConnectedException('Test NotConnectedException')



import os
import socket
import threading
from contextlib import ExitStack
from types import SimpleNamespace
from unittest.mock import MagicMock, mock_open, patch

import pytest

from pyx3270 import cli, server
from pyx3270.emulator import X3270, ExecutableApp

_real_socket_class = socket.socket


@pytest.fixture(autouse=True)
def no_os_exit(monkeypatch):
    def fake_exit(code=0):
        raise SystemExit(code)  # deixa o pytest lidar normalmente

    monkeypatch.setattr(os, '_exit', fake_exit)


@pytest.fixture
def mock_socket_with_accept():
    """
    Retorna uma tupla (mock_socket, mock_clientsock)
    mock_socket.accept() na primeira chamada retorna (mock_clientsock, addr),
    na segunda chamada levanta KeyboardInterrupt para parar o loop.
    """
    mock_socket = MagicMock()
    mock_clientsock = MagicMock()

    def accept_side_effect():
        yield (mock_clientsock, ('127.0.0.1', 12345))
        while True:
            raise KeyboardInterrupt('encerrar')

    mock_socket.accept.side_effect = accept_side_effect()
    return mock_socket, mock_clientsock


@pytest.fixture
def record_mocks():
    with ExitStack() as stack:
        # Patches no módulo 'pyx3270.server'
        connect_serversock = stack.enter_context(
            patch('pyx3270.server.connect_serversock')
        )
        ensure_dir = stack.enter_context(patch('pyx3270.server.ensure_dir'))
        is_screen = stack.enter_context(
            patch('pyx3270.server.is_screen_tn3270')
        )

        # Patches fora do server.py
        mock_select = stack.enter_context(patch('select.select'))
        mock_open_func = stack.enter_context(
            patch('builtins.open', new_callable=mock_open)
        )
        mock_join = stack.enter_context(patch('os.path.join'))

        mocks = SimpleNamespace(
            clientsock=MagicMock(spec=_real_socket_class, fileno=lambda: 3),
            serversock=MagicMock(spec=_real_socket_class, fileno=lambda: 4),
            connect_serversock=connect_serversock,
            ensure_dir=ensure_dir,
            is_screen=is_screen,
            select=mock_select,
            open_func=mock_open_func,
            join=mock_join,
        )

        yield mocks


@pytest.fixture
def mock_subprocess_popen(autouse=True):
    """Fixture para mockar subprocess.Popen."""
    with patch('subprocess.Popen') as mock_popen:
        mock_process = MagicMock()
        mock_process.stdin = MagicMock()
        mock_process.stdout = MagicMock()
        mock_process.stderr = MagicMock()
        mock_process.poll.return_value = None
        mock_process.pid = 12345
        mock_popen.return_value = mock_process
        yield mock_popen


@pytest.fixture
def mock_socket(autouse=True):
    """Fixture para mockar socket.socket e funções relacionadas."""
    with patch('socket.socket') as mock_socket_constructor:
        mock_sock_instance = MagicMock()
        mock_sock_instance.connect.return_value = None
        mock_sock_instance.makefile.return_value = MagicMock()
        mock_sock_instance.close.return_value = None
        mock_sock_instance.getsockname.return_value = ('127.0.0.1', 54321)
        mock_socket_constructor.return_value = mock_sock_instance
        with patch('socket.getaddrinfo'), patch('socket.gethostname'):
            yield mock_socket_constructor


@pytest.fixture
def mock_os_name(monkeypatch):
    """Fixture para mockar os.name."""
    monkeypatch.setattr('os.name', 'posix')


@pytest.fixture
def mock_executable_app_instance(mock_subprocess_popen):
    """Fixture para uma instância mockada de ExecutableApp."""
    with patch.object(
        ExecutableApp, '_spawn_app', return_value=None
    ), patch.object(
        ExecutableApp, '_get_executable_app_args', return_value=['dummy_args']
    ):
        app = ExecutableApp(shell=False, model='2')
        app.subprocess = mock_subprocess_popen.return_value
        app.subprocess.stdout.readline.return_value = b'ok\n'
        yield app


@pytest.fixture
def x3270_emulator_instance(mock_executable_app_instance):
    """Fixture para uma instância de X3270 com app mockado."""
    with patch.object(X3270, '_create_app', return_value=None):
        emulator = X3270(
            visible=False,
            model='3',
        )
        # Atribui manualmente o app mockado (já que _create_app está mockado)
        emulator.app = mock_executable_app_instance
        # Mocka o método principal de execução de comandos
        emulator._exec_command = MagicMock()
        yield emulator


@pytest.fixture
def x3270_real_exec_instance(mock_executable_app_instance):
    """Instância de X3270 com app mockado e Command mockado."""
    with patch.object(X3270, '_create_app', return_value=None):
        emulator = X3270(
            visible=False,
            model='3',
        )
        # Atribui manualmente o app mockado (já que _create_app está mockado)
        emulator.app = mock_executable_app_instance

        # Patch de Command no namespace do módulo pyx3270.emulator
        with patch('pyx3270.emulator.Command') as mock_command_class:
            mock_instance = MagicMock()
            mock_instance.execute.return_value = True
            mock_command_class.return_value = mock_instance

            yield emulator


@pytest.fixture
def x3270_cmd_instance(x3270_emulator_instance):
    """Fixture para uma instância de X3270Cmd associada a um X3270 mockado."""
    # Reseta o mock para cada teste
    x3270_emulator_instance._exec_command.reset_mock()
    # Retorna a instância do emulador que também é o Cmd
    return x3270_emulator_instance


@pytest.fixture
def record_dependencies():
    deps = SimpleNamespace(
        x3270=MagicMock(),
        server_thread=MagicMock(),
        address='localhost:3270',
        directory='./screens',
        tls=True,
        model='2',
        rich_print=MagicMock(),
        sys_exit=MagicMock(),
        control_replay=MagicMock(side_effect=KeyboardInterrupt),
    )

    # Patches globais dentro da fixture
    with patch('pyx3270.cli.X3270', return_value=deps.x3270), patch(
        'pyx3270.cli.start_server_thread', return_value=deps.server_thread
    ), patch('pyx3270.cli.control_replay', deps.control_replay), patch(
        'rich.print', deps.rich_print
    ):
        yield deps


@pytest.fixture
def replay_dependencies(tmp_path, monkeypatch):
    # Cria diretório e arquivo de teste para screens
    screens_dir = tmp_path / 'screens'
    screens_dir.mkdir()
    (screens_dir / '001.bin').write_bytes(
        b'\x11' * 200 + server.tn3270.IAC + server.tn3270.TN_EOR
    )

    # Configura valores padrões
    directory = str(screens_dir)
    port = 12345
    model = '2'
    tls = False

    # Mocks
    x3270_mock = type(
        'X3270Mock', (), {'connect_host': lambda self, *a, **kw: None}
    )()

    def control_replay_mock(th):
        # levanta KeyboardInterrupt quando chamado
        raise KeyboardInterrupt

    # Monkeypatch no módulo
    monkeypatch.setattr(cli, 'X3270', lambda *a, **kw: x3270_mock)
    monkeypatch.setattr(cli, 'start_command_process', lambda: None)
    monkeypatch.setattr(cli, 'control_replay', control_replay_mock)

    # Monkeypatch para não abrir socket de verdade
    def fake_start_server_thread(*args, **kwargs):
        thread = threading.Thread(target=lambda: None, daemon=True)
        thread.start()
        return thread

    monkeypatch.setattr(cli, 'start_server_thread', fake_start_server_thread)

    # Monkeypatch do rich.print para capturar chamadas
    printed_messages = []
    monkeypatch.setattr(
        cli.rich, 'print', lambda msg, **kwargs: printed_messages.append(msg)
    )

    # Retorna todos os valores úteis para o teste
    return dict(
        directory=directory,
        port=port,
        model=model,
        tls=tls,
        printed_messages=printed_messages,
        x3270_mock=x3270_mock,
    )



import os
import socket
from unittest.mock import MagicMock, Mock, call, mock_open, patch

import pytest

from pyx3270 import server, tn3270
from pyx3270.emulator import X3270
from pyx3270.server import ReplayState, process_command

_real_socket_class = socket.socket


def test_ensure_dir_exists(monkeypatch):
    """Testa ensure_dir quando o diretório já existe."""
    mock_isdir = MagicMock(return_value=True)
    mock_makedirs = MagicMock()
    monkeypatch.setattr(os.path, 'isdir', mock_isdir)
    monkeypatch.setattr(os, 'makedirs', mock_makedirs)

    server.ensure_dir('/path/to/existing_dir')

    mock_isdir.assert_called_once_with('/path/to/existing_dir')
    mock_makedirs.assert_not_called()


def test_ensure_dir_creates(monkeypatch):
    """Testa ensure_dir quando o diretório não existe."""
    mock_isdir = MagicMock(return_value=False)
    mock_makedirs = MagicMock()
    monkeypatch.setattr(os.path, 'isdir', mock_isdir)
    monkeypatch.setattr(os, 'makedirs', mock_makedirs)

    server.ensure_dir('/path/to/new_dir')

    mock_isdir.assert_called_once_with('/path/to/new_dir')
    mock_makedirs.assert_called_once_with('/path/to/new_dir')


def test_ensure_dir_no_path():
    """Testa ensure_dir com path vazio ou None."""
    # Não deve fazer nada nem levantar erro
    server.ensure_dir(None)
    server.ensure_dir('')


def test_is_screen_tn3270():
    """Testa a detecção de telas TN3270."""
    assert (
        server.is_screen_tn3270(b'\x00' * 99 + b'\x11') is False
    )  # Muito curto
    assert (
        server.is_screen_tn3270(b'\x00' * 101) is False
    )  # Sem byte de comando
    assert (
        server.is_screen_tn3270(b'\x00' * 100 + b'\x11') is True
    )  # Tamanho e byte OK
    assert (
        server.is_screen_tn3270(b'\x11' + b'\x00' * 100) is True
    )  # Byte no início


def test_load_screens_empty_dir(tmp_path, monkeypatch):
    """Testa load_screens com diretório vazio."""
    record_dir = tmp_path / 'records'
    record_dir.mkdir()
    mock_ensure_dir = MagicMock()
    monkeypatch.setattr(server, 'ensure_dir', mock_ensure_dir)

    screens = server.load_screens_basic(str(record_dir))
    mock_ensure_dir.assert_called_once_with(str(record_dir))
    assert screens == dict()


def test_load_screens_with_files(tmp_path, monkeypatch):
    """Testa load_screens com arquivos .bin."""
    record_dir = tmp_path / 'records'
    record_dir.mkdir()
    # Cria arquivos de teste
    (record_dir / '001.bin').write_bytes(
        b'screen1_data' + tn3270.IAC + tn3270.TN_EOR
    )
    (record_dir / '000.bin').write_bytes(b'screen0_data')  # Sem EOR
    (record_dir / 'ignored.txt').write_text('ignore me')
    (record_dir / '002.bin').write_bytes(
        b'screen2_data' + tn3270.IAC + tn3270.TN_EOR
    )

    mock_ensure_dir = MagicMock()
    monkeypatch.setattr(server, 'ensure_dir', mock_ensure_dir)

    screens = server.load_screens_basic(str(record_dir))
    screens_list = list(screens.values())

    mock_ensure_dir.assert_called_once_with(str(record_dir))
    EXPECTED_FILES = {'000', '001', '002'}
    assert set(screens.keys()) == EXPECTED_FILES
    assert len(screens_list) == len(EXPECTED_FILES)
    # Verifica a ordem e o conteúdo (com EOR adicionado se necessário)
    assert screens_list[0] == b'screen0_data' + tn3270.IAC + tn3270.TN_EOR
    assert screens_list[1] == b'screen1_data' + tn3270.IAC + tn3270.TN_EOR
    assert screens_list[2] == b'screen2_data' + tn3270.IAC + tn3270.TN_EOR


def test_convert_s():
    """Testa a conversão de string hexadecimal com SF."""
    assert (
        server.convert_s('11C1C1 SF(C1=01, C2=02) 11C2C2')
        == '11C1C11D011D0211C2C2'
    )
    assert server.convert_s('SF(A=1, B=2, C=3)') == '1D11D21D3'
    assert server.convert_s('No SF here') == 'NoSFhere'
    assert server.convert_s('SF() Empty') == 'Empty'  # SF vazio é removido
    assert (
        server.convert_s('SF(X=10) 1234 SF(Y=20, Z=30)') == '1D1012341D201D30'
    )


@patch('socket.create_connection', autospec=True)
def test_connect_serversock_success(mock_create_conn):
    """Testa connect_serversock com sucesso."""
    mock_client_sock = MagicMock()
    mock_server_sock = MagicMock()
    mock_create_conn.return_value = mock_server_sock

    result_sock = server.connect_serversock(
        mock_client_sock, 'mainframe.com:3270'
    )

    mock_create_conn.assert_called_once_with(
        ('mainframe.com', 3270), timeout=5
    )
    assert result_sock == mock_server_sock
    mock_client_sock.close.assert_not_called()


@patch('socket.create_connection', autospec=True)
def test_connect_serversock_default_port(mock_create_conn):
    """Testa connect_serversock usando a porta padrão."""
    mock_client_sock = MagicMock()
    mock_server_sock = MagicMock()
    mock_create_conn.return_value = mock_server_sock

    result_sock = server.connect_serversock(mock_client_sock, 'onlyhost')

    mock_create_conn.assert_called_once_with(('onlyhost', 3270), timeout=5)
    assert result_sock == mock_server_sock


@patch('socket.create_connection', autospec=True)
def test_connect_serversock_failure(mock_create_conn):
    """Testa connect_serversock com falha na conexão."""
    mock_client_sock = MagicMock()
    mock_create_conn.side_effect = socket.timeout('Connection timed out')

    result_sock = server.connect_serversock(mock_client_sock, 'badhost:1234')

    mock_create_conn.assert_called_once_with(('badhost', 1234), timeout=5)
    assert result_sock is None
    mock_client_sock.close.assert_called_once()


@patch('socket.socket')
def test_backend_3270_navigation(mock_socket_constructor):
    """Testa a navegação básica em backend_3270."""
    mock_clientsock = MagicMock(spec=_real_socket_class)
    # Simula recebimento de AID e dados de tecla pressionada
    mock_clientsock.recv.side_effect = [
        tn3270.ENTER,
        b'K\xe9\xff',  # Enter pressionado
        tn3270.PF3,
        b'K\xe9\xff',  # PF3 pressionado
        tn3270.PF8,
        b'K\xe9\xff',  # PF8 pressionado
        tn3270.CLEAR,
        b'K\xe9\xff',  # Clear pressionado
        tn3270.PF7,
        b'K\xe9\xff',  # PF7 pressionado (emulator=True)
        tn3270.PF4,
        b'K\xe9\xff',  # PF4 pressionado
        b'\x00',  # Simula fechamento do terminal para sair do loop
    ]
    screens = {'s0': b's0', 's1': b's1', 's2': b's2', 's3': b's3'}

    # Teste 1: ENTER
    result = server.backend_3270(mock_clientsock, screens, 0, emulator=True)
    assert result == {'current_screen': 1, 'clear': False}

    # Teste 2: PF3
    result = server.backend_3270(mock_clientsock, screens, 1, emulator=True)
    assert result == {'current_screen': 0, 'clear': False}

    # Teste 3: PF8
    result = server.backend_3270(mock_clientsock, screens, 0, emulator=True)
    assert result == {'current_screen': 1, 'clear': False}

    # Teste 4: CLEAR
    result = server.backend_3270(mock_clientsock, screens, 1, emulator=True)
    assert result == {'current_screen': 1, 'clear': True}
    mock_clientsock.sendall.assert_called_with(tn3270.CLEAR_SCREEN_BUFFER)

    # Teste 5: PF7 (emulator=True)
    result = server.backend_3270(mock_clientsock, screens, 2, emulator=True)
    assert result == {'current_screen': 1, 'clear': False}

    # Teste 6: PF4
    result = server.backend_3270(mock_clientsock, screens, 1, emulator=True)
    assert result == {'current_screen': 2, 'clear': False}

    # Teste 7: Fechamento do terminal
    mock_clientsock.recv.side_effect = [b'']
    with pytest.raises(ConnectionResetError, match='Terminal fechado.'):
        server.backend_3270(mock_clientsock, screens, 2, emulator=True)


@patch('socket.socket')
def test_backend_3270_timeout(mock_socket_constructor):
    """Testa backend_3270 com timeout no recv."""
    EXPECTED_CALLS = 3
    mock_clientsock = MagicMock(spec=_real_socket_class)
    mock_clientsock.recv.side_effect = [
        socket.timeout,
        tn3270.ENTER,
        b'K\xe9\xff',
    ]
    screens = {'s0': b's0', 's1': b's1'}

    result = server.backend_3270(mock_clientsock, screens, 0, emulator=False)
    assert result == {'current_screen': 0, 'clear': False}
    assert mock_clientsock.recv.call_count == EXPECTED_CALLS


@patch('pyx3270.server.backend_3270')
def test_replay_handler(mock_backend, monkeypatch):
    """Testa o fluxo básico de replay_handler."""
    mock_clientsock = MagicMock(spec=_real_socket_class)
    screens = {
        'screen0': b'screen0',
        'screen1': b'screen1',
        'screen2': b'screen2',
    }
    screens_list = list(screens.values())
    # Simula a sequência de interações retornada por backend_3270
    mock_backend.side_effect = [
        {'current_screen': 0, 'clear': False},  # Inicial
        {'current_screen': 1, 'clear': False},  # Após interação 1
        {'current_screen': 1, 'clear': True},  # Após interação 2 (Clear)
        {'current_screen': 2, 'clear': False},  # Após interação 3
        Exception('Simulate loop break'),  # Para sair do loop
    ]

    server.replay_handler(
        mock_clientsock, screens, emulator=True, base_directory='./screens'
    )

    # Verifica chamadas
    mock_clientsock.sendall.assert_has_calls([
        call(b'\xff\xfd\x18\xff\xfb\x18'),  # handshake inicial
        call(screens_list[0]),  # inicial
        call(screens_list[0]),  # chamada extra repetida
        call(screens_list[1]),
        call(screens_list[2]),
    ])
    EXPECTED_CALLS = 5
    assert mock_backend.call_count == EXPECTED_CALLS  # Chamado até a exceção
    mock_clientsock.close.assert_called_once()  # Garante que fechou no finally


def test_record_handler_basic_flow(record_mocks):
    """Testa o fluxo básico de gravação em record_handler (sem TLS)."""
    mock_clientsock = record_mocks.clientsock
    mock_serversock = record_mocks.serversock
    record_mocks.connect_serversock.return_value = mock_serversock
    mock_emu = MagicMock(spec=X3270)
    mock_emu.tls = False  # Teste sem TLS
    record_dir = '/fake/dir'
    record_mocks.join.side_effect = lambda *args: os.path.normpath(
        '/'.join(args)
    )

    # Dados simulados recebidos
    client_data = b'client_req' + tn3270.IAC + tn3270.TN_EOR
    server_data_screen1 = b'server_resp_screen1' + tn3270.IAC + tn3270.TN_EOR
    server_data_non_screen = b'server_non_screen_data'
    server_data_screen2 = b'server_resp_screen2' + tn3270.IAC + tn3270.TN_EOR

    # Configura select para retornar sockets e simular fim
    record_mocks.select.side_effect = [
        ([mock_clientsock], [], []),  # Cliente envia
        ([mock_serversock], [], []),  # Servidor envia tela 1
        ([mock_serversock], [], []),  # Servidor envia dados não-tela
        ([mock_serversock], [], []),  # Servidor envia tela 2
        ConnectionResetError,  # Simula desconexão para parar o loop
    ]

    # Configura recv dos sockets
    mock_clientsock.recv.return_value = client_data
    mock_serversock.recv.side_effect = [
        server_data_screen1,
        server_data_non_screen,
        server_data_screen2,
    ]

    # Configura is_screen_tn3270
    record_mocks.is_screen.side_effect = (
        lambda data: tn3270.IAC + tn3270.TN_EOR in data
    )

    server.record_handler(mock_clientsock, mock_emu, 'host:3270', record_dir)

    # Verificações
    record_mocks.connect_serversock.assert_called_once_with(
        mock_clientsock, 'host:3270'
    )

    record_mocks.ensure_dir.assert_called_once_with(record_dir)
    EXPECTED_CALLS = 5
    assert record_mocks.select.call_count == EXPECTED_CALLS

    # Verifica envios entre sockets
    mock_serversock.sendall.assert_called_once_with(client_data)
    mock_clientsock.sendall.assert_has_calls([
        call(server_data_screen1),
        call(server_data_non_screen),
        call(server_data_screen2),
    ])

    # Verifica gravação dos arquivos
    # Chamado para tela 1 e tela 2
    EXPECTED_CALLS = 2
    assert record_mocks.join.call_count == EXPECTED_CALLS
    record_mocks.open_func.assert_has_calls(
        [
            call(os.path.normpath('/fake/dir/000.bin'), 'wb'),
            call(os.path.normpath('/fake/dir/001.bin'), 'wb'),
        ],
        any_order=True,
    )

    # Verifica o conteúdo escrito (mock_open captura escritas)
    handle = (
        record_mocks.open_func()
    )  # mock do arquivo aberto (mesmo para todos)

    # Obtem todas as chamadas ao write (argumentos escritos)
    write_calls = [
        call_args[0][0] for call_args in handle.write.call_args_list
    ]

    assert any(server_data_screen1 in w for w in write_calls), (
        'server_data_screen1 não foi escrito'
    )
    assert any(server_data_screen2 in w for w in write_calls), (
        'server_data_screen2 não foi escrito'
    )

    # Verifica fechamento dos sockets
    mock_clientsock.close.assert_called_once()
    mock_serversock.close.assert_called_once()


@patch('pyx3270.server.connect_serversock')
def test_record_handler_connect_fail(mock_connect_serversock):
    """Testa record_handler quando a conexão inicial falha."""
    mock_clientsock = MagicMock(spec=_real_socket_class)
    mock_connect_serversock.return_value = None  # Simula falha
    mock_emu = MagicMock(spec=X3270)

    server.record_handler(mock_clientsock, mock_emu, 'badhost:port', '/dir')

    mock_connect_serversock.assert_called_once_with(
        mock_clientsock, 'badhost:port'
    )
    # Não deve tentar fechar sockets ou fazer outras operações
    mock_clientsock.close.assert_not_called()


def test_record_handler_not_is_screen_tn3270():
    fake_client_sock = MagicMock()
    fake_server_sock = MagicMock()

    # Simula dados vindos do servidor com terminador TN_EOR
    tn_eor = b'\xff\xef'  # IAC + TN_EOR
    fake_data = b'nao eh tn3270' + tn_eor

    # recv() devolve os dados na primeira chamada e depois vazio
    fake_server_sock.recv.side_effect = [fake_data, b'']
    fake_client_sock.recv.side_effect = [
        b'',
        b'',
    ]  # Não recebe nada do cliente

    # select retorna sockets prontos para leitura
    def fake_select(rlist, _, __, ___):
        if fake_server_sock in rlist:
            return ([fake_server_sock], [], [])
        return ([], [], [])

    with patch(
        'pyx3270.server.is_screen_tn3270', return_value=False
    ) as mock_is_screen, patch(
        'pyx3270.server.connect_serversock', return_value=fake_server_sock
    ), patch('select.select', side_effect=fake_select):
        emu = MagicMock(spec=X3270)
        emu.tls = False

        server.record_handler(
            clientsock=fake_client_sock,
            emu=emu,
            address='127.0.0.1',
            record_dir='/screens',
            delay=0.01,
        )

        # Garante que is_screen_tn3270 foi chamado
        assert mock_is_screen.called


def test_load_screens_calls_ensure_dir(monkeypatch, tmp_path):
    record_dir = str(tmp_path)

    mock_ensure_dir = MagicMock()
    monkeypatch.setattr(server, 'ensure_dir', mock_ensure_dir)
    # Forçar que haja arquivos .bin para evitar o fallback
    monkeypatch.setattr(
        server.os, 'listdir', MagicMock(return_value=['file.bin'])
    )
    monkeypatch.setattr(
        server.os,
        'walk',
        MagicMock(return_value=[(record_dir, [], ['file.bin'])]),
    )
    monkeypatch.setattr(
        server, 'tn3270', MagicMock(IAC=b'\xff', TN_EOR=b'\xef')
    )

    open_mock = MagicMock()
    open_mock.return_value.__enter__.return_value.read.return_value = (
        b'data' + server.tn3270.IAC + server.tn3270.TN_EOR
    )
    m = mock_open(read_data=b'data' + server.tn3270.IAC + server.tn3270.TN_EOR)
    monkeypatch.setattr('builtins.open', m)

    server.load_screens(record_dir)

    mock_ensure_dir.assert_called_once_with(record_dir)


def test_load_screens_fallback_to_load_screens_basic(monkeypatch):
    record_dir = '/fake/dir'

    monkeypatch.setattr(server, 'ensure_dir', MagicMock())
    monkeypatch.setattr(
        server.os, 'listdir', MagicMock(return_value=['file.txt', 'other.dat'])
    )
    mock_load_basic = MagicMock(return_value={'basic': 'screens'})
    monkeypatch.setattr(server, 'load_screens_basic', mock_load_basic)

    result = server.load_screens(record_dir)

    mock_load_basic.assert_called_once_with(server.BINARY_FOLDER)
    assert result == {'basic': 'screens'}


def test_load_screens_logs_error_and_returns_empty(monkeypatch):
    record_dir = '/fake/dir'

    # Forçar ensure_dir sem efeitos
    monkeypatch.setattr(server, 'ensure_dir', MagicMock())

    # Forçar que listdir lance exceção para simular erro
    def raise_os_error(_):
        raise OSError('fail')

    monkeypatch.setattr(server.os, 'listdir', raise_os_error)

    mock_logger = MagicMock()
    monkeypatch.setattr(server, 'logger', mock_logger)

    result = server.load_screens(record_dir)

    mock_logger.error.assert_called_once_with(
        f'Falha ao carregar caminho: {record_dir}'
    )
    assert result == {}


def test_find_directory_finds_match():
    base_dir = '/fake/base'
    search_name = 'testdir'

    # Simula listdir retornando vários nomes
    mock_listdir = ['OtherDir', 'TestDir123', 'anotherdir']

    # Simula isdir para retornar True apenas para 'TestDir123' e 'OtherDir'
    def mock_isdir(path):
        return path.endswith('TestDir123') or path.endswith('OtherDir')

    with patch('os.listdir', return_value=mock_listdir), patch(
        'os.path.isdir', side_effect=mock_isdir
    ):
        found = server.find_directory(base_dir, search_name)

    expected_path = os.path.join(base_dir, 'TestDir123')
    assert found == expected_path


def test_find_directory_skips_non_dirs():
    base_dir = '/fake/base'
    search_name = 'nomatch'

    mock_listdir = ['file.txt', 'notadir', 'maybeadir']

    # Só 'maybeadir' é diretório
    def mock_isdir(path):
        return path.endswith('maybeadir')

    with patch('os.listdir', return_value=mock_listdir), patch(
        'os.path.isdir', side_effect=mock_isdir
    ):
        found = server.find_directory(base_dir, search_name)

    # Não encontrou, deve retornar None
    assert found is None


def test_find_directory_returns_none_when_no_match():
    base_dir = '/fake/base'
    search_name = 'absent'

    mock_listdir = ['DirOne', 'DirTwo']

    # Todos são diretórios
    def mock_isdir(path):
        return True

    with patch('os.listdir', return_value=mock_listdir), patch(
        'os.path.isdir', side_effect=mock_isdir
    ):
        found = server.find_directory(base_dir, search_name)

    assert found is None


def test_handle_set_found():
    screens = {
        'SCREEN1': b'data1',
        'SCREEN2': b'data2',
        'OTHER': b'data3',
    }
    command = 'set screen2'

    result = server.handle_set(command, screens)
    assert result == 1  # índice de 'SCREEN2'


def test_handle_set_not_found():
    screens = {
        'SCREEN1': b'data1',
        'SCREEN2': b'data2',
    }
    command = 'set nonexistent'

    result = server.handle_set(command, screens)
    assert result is None


@patch('pyx3270.server.logger')
def test_handle_add_valid(mock_logger):
    screens = {}
    screens_list = []

    # Compose a hex string for simple bytes 'ABC' (hex 41 42 43)
    hex_data = '414243'
    command = f'add newscreen {hex_data}'

    server.handle_add(command, screens, screens_list)

    expected_bytes = (
        tn3270.START_SCREEN
        + bytes.fromhex(hex_data)
        + tn3270.IAC
        + tn3270.TN_EOR
    )

    # Verifica se a tela foi adicionada no dict e na lista
    assert 'NEWSCREEN' in screens
    assert screens['NEWSCREEN'] == expected_bytes
    assert expected_bytes in screens_list

    # Logger warning não deve ser chamado
    mock_logger.warning.assert_not_called()


@patch('pyx3270.server.logger')
def test_handle_add_invalid_format(mock_logger):
    screens = {}
    screens_list = []

    # Comando com formato inválido (falta argumento)
    command = 'add invalidcommand'

    server.handle_add(command, screens, screens_list)

    # Nenhuma tela deve ter sido adicionada
    assert screens == {}
    assert screens_list == []

    # Logger deve ter sido chamado com warning
    mock_logger.warning.assert_called_once_with(
        '[!] Formato inválido ao adicionar tela'
    )


@pytest.mark.parametrize(
    ('command', 'expected_screen', 'expected_clear', 'should_close'),
    [
        ('set TELA', 1, False, False),
        ('add TELA 00ff', 0, False, False),
        ('change directory teste', 0, False, False),
        ('next', 1, False, False),
        ('prev', 0, False, False),
        ('clear', 0, True, False),
        ('q', 0, False, True),
        ('comando_invalido', 0, False, False),
    ],
)
def test_process_command(
    command, expected_screen, expected_clear, should_close
):
    # Setup mocks e estado
    mock_socket = Mock()

    mock_screens = {'TELA_PRINCIPAL': b'tela1', 'OUTRA_TELA': b'tela2'}
    # Dados fictícios para inicialização
    screens = {'TELA': 'conteúdo fictício'}
    screens_list = ['TELA']
    current_screen = 0
    clear = False

    state = ReplayState(screens, screens_list, current_screen, clear)
    state.screens = mock_screens
    state.screens_list = list(mock_screens.values())
    state.current_screen = 0
    state.clear = False

    with patch('pyx3270.server.handle_set', return_value=1) as mock_set, patch(
        'pyx3270.server.handle_add'
    ) as mock_add, patch(
        'pyx3270.server.handle_change_directory',
        return_value=(mock_screens, list(mock_screens.values())),
    ) as mock_change:
        result_state = process_command(command, mock_socket, 'base', state)

        # Checagens
        assert result_state.current_screen == expected_screen
        assert result_state.clear == expected_clear

        if should_close:
            mock_socket.close.assert_called_once()
        else:
            mock_socket.close.assert_not_called()

        if command == 'clear':
            mock_socket.sendall.assert_called_once_with(
                tn3270.CLEAR_SCREEN_BUFFER
            )
        else:
            mock_socket.sendall.assert_not_called()

        if command.startswith('set '):
            mock_set.assert_called_once()
        else:
            mock_set.assert_not_called()

        if command.startswith('add '):
            mock_add.assert_called_once()
        else:
            mock_add.assert_not_called()

        if command.startswith('change directory '):
            mock_change.assert_called_once()
        else:
            mock_change.assert_not_called()


def test_invalid_handle_change_directory():
    """Testa handle_change_directory com diretório inválido."""
    command = 'change directory /invalid/path'

    with patch('pyx3270.server.logger'):
        screens_result, screens_list = server.handle_change_directory(
            command, './screens'
        )

        assert screens_result == dict()
        assert screens_list == list()



import functools
import warnings
from typing import Any


def x3270_builtins_class(cls):
    """Decorador de classe que aplica x3270_builtins a todos os métodos."""
    for name, attr in cls.__dict__.items():
        if callable(attr) and not name.startswith('__'):
            # aplica o wrapper
            setattr(cls, name, _wrap_method(attr))
    return cls


def _wrap_method(method):
    """Wrapper que imprime o nome do método e delega para _target."""

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        return x3270_command(self, method.__name__, *args, **kwargs)

    return wrapper


def x3270_command(em, func, *args, **kwargs):
    all_args_str = ', '.join(
        list(map(str, args)) + [f'{k}={repr(v)}' for k, v in kwargs.items()]
    )

    if func == 'send_string_not_log':
        warnings.warn(
            f'`{func}` foi descontinuada, use `send_string`'
            f' com argumento `password=True`.',
            DeprecationWarning,
        )
        return em.send_string(*args, **kwargs, password=True)

    if 'send_pf' in func or func == 'pf':
        if not all_args_str:
            warnings.warn(
                f'`{func}` foi descontinuada, use `send_pf` com argumento.',
                DeprecationWarning,
            )
            all_args_str = func[-1]

        em._exec_command(f'PF({all_args_str})'.encode('utf8'))
        em.wait(em.time_unlock, 'unlock')
        return

    if func == 'connect' and (len(args) + len(kwargs)) > 1:
        return em.connect_host(*args, **kwargs)

    try:
        cmd = em._exec_command(f'{func}({all_args_str})'.encode('utf8'))
        try:
            text = [text.decode('utf8') for text in cmd.data[0:]]
            result = ' '.join(text)
        except AttributeError:
            result = [val for val in cmd.data[0:]]

        return result
    except Exception:
        raise


@x3270_builtins_class
class X3270Commands:
    def abort(self) -> Any: ...  # Aborta scripts e macros pendentes.
    def ansitext(self) -> Any: ...  # Exibe texto NVT pendente.
    def ascii(self, *args: Any) -> Any: ...  # Conteúdo da tela em ASCII.
    def ascii1(
        self, row: int, col: int
    ) -> Any: ...  # Bytes da tela em ASCII da posição (1-origem).
    def asciifield(self) -> Any: ...  # Conteúdo do campo atual em ASCII.
    def attn(self) -> Any: ...  # Envia a sequência ATTN do 3270.
    def backspace(
        self,
    ) -> Any: ...  # Move o cursor uma posição para a esquerda.
    def backtab(self) -> Any: ...  # Move o cursor para o campo anterior.
    def bell(self) -> Any: ...  # Toca o sino do terminal.
    def circumnot(self) -> Any: ...  # Envia ~ em NVT ou ¬ em 3270.
    def clear(self) -> Any: ...  # Limpa a tela do terminal 3270.
    def close(self) -> Any: ...  # Alias para Disconnect, fecha a conexão.
    def closescript(self) -> Any: ...  # Encerra o script em execução.
    def compose(
        self,
    ) -> Any: ...  # Interpreta as próximas duas teclas conforme o mapa.
    def connect(
        self, host: str, port: str | None = None
    ) -> Any: ...  # Conecta ao host.
    def cursorselect(
        self,
    ) -> Any: ...  # Seleciona local do cursor como caneta luminosa.
    def delete(self) -> Any: ...  # Apaga caractere na posição do cursor.
    def deletefield(self) -> Any: ...  # Apaga o conteúdo do campo atual.
    def deleteword(self) -> Any: ...  # Apaga a palavra anterior ao cursor.
    def disconnect(self) -> Any: ...  # Fecha conexão com o host.
    def down(self) -> Any: ...  # Move o cursor para a linha de baixo.
    def dup(self) -> Any: ...  # Envia tecla DUP do 3270.
    def ebcdic(self) -> Any: ...  # Conteúdo da tela em EBCDIC.
    def ebcdicfield(self) -> Any: ...  # Conteúdo do campo atual em EBCDIC.
    def enter(self) -> Any: ...  # Envia o comando ENTER.
    def erase(self) -> Any: ...  # Backspace destrutivo.
    def eraseeof(self) -> Any: ...  # Apaga até o fim do campo.
    def eraseinput(self) -> Any: ...  # Apaga todos os campos de entrada.
    def escape(self) -> Any: ...  # Abre prompt c3270>
    def execute(self, command: str) -> Any: ...  # Executa comando no shell.
    def exit(self) -> Any: ...  # Sai do c3270.
    def expect(self, pattern: str) -> Any: ...  # Aguarda saída NVT.
    def fieldend(self) -> Any: ...  # Move cursor ao fim do campo.
    def fieldmark(self) -> Any: ...  # Envia tecla FIELD MARK.
    def flip(self) -> Any: ...  # Inverte a tela (espelho).
    def help(self, action: str) -> Any: ...  # Exibe ajuda para um tópico.
    def hexstring(self, digits: str) -> Any: ...  # Envia dados em hexadecimal.
    def home(self) -> Any: ...  # Move o cursor para o primeiro campo.
    def ignore(self) -> Any: ...  # Não faz nada.
    def info(
        self, text: str
    ) -> Any: ...  # Exibe texto na barra de status (OIA).
    def insert(self) -> Any: ...  # Ativa modo de inserção 3270.
    def interrupt(self) -> Any: ...  # Envia comando TELNET IAC IP.
    def key(self, symbol: str) -> Any: ...  # Envia um caractere específico.
    def keyboarddisable(
        self, mode: str
    ) -> Any: ...  # Modifica o bloqueio automático.
    def keymap(self, keymap_name: str) -> Any: ...  # Ativa keymap temporário.
    def keypad(self) -> Any: ...  # Mostra o teclado 3270 na tela.
    def left(self) -> Any: ...  # Move o cursor 1 coluna à esquerda.
    def left2(self) -> Any: ...  # Move o cursor 2 colunas à esquerda.
    def macro(self, name: str) -> Any: ...  # Executa um macro definido.
    def menu(self) -> Any: ...  # Exibe o menu de comandos.
    def movecursor(
        self, row: int, col: int
    ) -> Any: ...  # Move cursor para linha e coluna (origem 0).
    def movecursor1(
        self, row: int, col: int
    ) -> Any: ...  # Move cursor para linha e coluna (origem 1).
    def movecursoroffset(
        self, offset: int
    ) -> Any: ...  # Move cursor para offset na memória.
    def newline(
        self,
    ) -> Any: ...  # Move cursor para primeiro campo da linha seguinte.
    def nextword(self) -> Any: ...  # Move cursor para a próxima palavra.
    def previousword(self) -> Any: ...  # Move cursor para a palavra anterior.
    def open(self) -> Any: ...  # Alias para Connect.
    def pa(self, n: int) -> Any: ...  # Envia tecla PA1, PA2, etc.
    def pf(self, n: int) -> Any: ...  # Envia tecla PF1 a PF24.
    def pause(self) -> Any: ...  # Espera 350ms.
    def paste(self) -> Any: ...  # Cola conteúdo do clipboard.
    def printer(
        self, start: bool, lu: str, stop: bool
    ) -> Any: ...  # Inicia ou para sessão de impressão 3287.
    def printtext(
        self, file: str
    ) -> Any: ...  # Salva ou imprime imagem da tela.
    def prompt(self, app_name: str) -> Any: ...  # Abre prompt externo.
    def query(
        self, keyword: str
    ) -> Any: ...  # Consulta parâmetros operacionais.
    def quit(self) -> Any: ...  # Sai do terminal 3270.
    def readbuffer(self, mode: str) -> Any: ...  # Despeja o buffer de tela.
    def reconnect(self) -> Any: ...  # Reconecta ao host anterior.
    def redraw(self) -> Any: ...  # Reexibe a tela do terminal.
    def reset(self) -> Any: ...  # Desbloqueia o teclado.
    def right(self) -> Any: ...  # Move cursor 1 coluna à direita.
    def right2(self) -> Any: ...  # Move cursor 2 colunas à direita.
    def screentrace(
        self, mode: str, file: str
    ) -> Any: ...  # Salva imagens da tela em arquivo.
    def script(self, path: str) -> Any: ...  # Executa script filho.
    def scroll(self, direction: str) -> Any: ...  # Rola a tela.
    def set(self) -> Any: ...  # Altera ou exibe configurações.
    def show(self, item: str) -> Any: ...  # Exibe status e configurações.
    def snap(self, *args: str) -> Any: ...  # Manipula snapshots da tela.
    def source(self, file: str) -> Any: ...  # Executa comandos de um arquivo.
    def string(self, text: str) -> Any: ...  # Envia string para o campo atual.
    def sysreq(self) -> Any: ...  # Envia tecla System Request (SysReq).
    def tab(self) -> Any: ...  # Move cursor para próximo campo.
    def temporarycomposemap(
        self, map: str
    ) -> Any: ...  # Define mapa de composição temporário.
    def temporarykeymap(
        self, keymap_name: str
    ) -> Any: ...  # Alias para Keymap.
    def toggle(
        self, name: str, value: str
    ) -> Any: ...  # Alterna configuração específica.
    def toggleinsert(self) -> Any: ...  # Ativa/desativa modo inserção.
    def togglereverse(
        self,
    ) -> Any: ...  # Ativa/desativa modo reverso de entrada.
    def trace(self, mode: str) -> Any: ...  # Configura rastreamento.
    def transfer(selfs: str) -> Any: ...  # Transferência de arquivos IND$FILE.
    def up(self) -> Any: ...  # Move o cursor para cima.
    def wait(self, *args: Any) -> Any: ...  # Aguarda eventos do host.



from abc import ABC, abstractmethod

from pyx3270.x3270_commands import X3270Commands


class AbstractExecutableApp(ABC):
    """Representa uma aplicação responsavel por emular um terminal tn3270."""

    @classmethod
    @abstractmethod
    def connect(*args) -> bool: ...

    @abstractmethod
    def close(self) -> None: ...

    @abstractmethod
    def write(self, data: str) -> None: ...

    @abstractmethod
    def readline(self) -> bytes: ...

    @abstractmethod
    def _spawn_app(self) -> None: ...

    @abstractmethod
    def _get_executable_app_args(self, model: str) -> None: ...


class AbstractCommand(ABC):
    """Representa um comando de script no TerminalClient."""

    @abstractmethod
    def execute(self) -> bool: ...

    @abstractmethod
    def handle_result(self, result: str) -> bool: ...


class AbstractEmulatorCmd(ABC, X3270Commands):
    @abstractmethod
    def clear_screen(self) -> None: ...

    @abstractmethod
    def wait_for_field(self, timeout: float) -> None: ...

    @abstractmethod
    def wait_string_found(
        self, ypos: int, xpos: int, string: str, equal: bool, timeout: float
    ) -> bool: ...

    @abstractmethod
    def string_found(self, ypos: int, xpos: int, string: str) -> bool: ...

    @abstractmethod
    def delete_field(self) -> None: ...

    @abstractmethod
    def move_to(self, ypos: int, xpos: int) -> None: ...

    @abstractmethod
    def send_pf(self, value: str) -> None: ...

    @abstractmethod
    def send_string(self, tosend: str, ypos: int, xpos: int) -> None: ...

    @abstractmethod
    def send_enter(self, times: int = 1) -> None: ...

    @abstractmethod
    def send_home(self) -> None: ...

    @abstractmethod
    def get_string(self) -> str: ...

    @abstractmethod
    def get_string_area(
        self, yposi: int, xposi: int, ypose: int, xpose: int
    ) -> str: ...

    @abstractmethod
    def get_full_screen(self, header: bool) -> str: ...

    @abstractmethod
    def save_screen(self, file_path: str, file_name: str) -> None: ...

    @abstractmethod
    def search_string(
        self, string: str, ignore_case: bool = False
    ) -> bool: ...

    @abstractmethod
    def get_string_positions(
        self, string: str, ignore_case: bool = False
    ) -> list[tuple[int, int]]: ...

    @abstractmethod
    def _get_ypos_and_xpos_from_index(self, index: int) -> tuple[int, int]: ...


class AbstractEmulator(AbstractEmulatorCmd):
    """
    Representa um subprocesso do emulador x/s3270,
    fornece uma API para interagir com ele.
    """

    @abstractmethod
    def terminate(self) -> None: ...

    @abstractmethod
    def is_connected(self) -> bool: ...

    @abstractmethod
    def connect_host(self, host: str, port: str, tls: bool) -> None: ...

    @abstractmethod
    def reconnect_host(self) -> None: ...

    @abstractmethod
    def _create_app(self) -> None: ...

    @abstractmethod
    def _exec_command(self, cmdstr: str) -> str: ...



import os

from PyInstaller.utils.hooks import get_package_paths

# Caminho do pacote instalado
package_path, _ = get_package_paths('pyx3270')

# Caminho da pasta bin dentro do pacote
bin_dir = os.path.join(package_path, 'pyx3270', 'bin')

datas = []

if os.path.isdir(bin_dir):
    for root, _, files in os.walk(bin_dir):
        for file in files:
            full_path = os.path.join(root, file)
            # Caminho relativo a partir da pasta bin (ex: windows/wc3270.exe)
            rel_path = os.path.relpath(full_path, bin_dir)
            # Corrige o destino para ser uma pasta, não um arquivo
            target_dir = os.path.join(
                'pyx3270', 'bin', os.path.dirname(rel_path)
            )
            datas.append((full_path, target_dir))

__all__ = ['datas']



LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        }
    },
    'handlers': {
        'emulator_file': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'x3270_emulator.log',
            'when': 'midnight',
            'interval': 1,
            'backupCount': 7,
            'encoding': 'utf-8',
            'formatter': 'default',
        },
        'server_file': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'x3270_server.log',
            'when': 'midnight',
            'interval': 1,
            'backupCount': 7,
            'encoding': 'utf-8',
            'formatter': 'default',
        },
        'offline_file': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'x3270_offline.log',
            'when': 'midnight',
            'interval': 1,
            'backupCount': 7,
            'encoding': 'utf-8',
            'formatter': 'default',
        },
    },
    'loggers': {
        'pyx3270.emulator': {
            'level': 'DEBUG',
            'handlers': ['emulator_file'],
            'propagate': False,
        },
        'pyx3270.server': {
            'level': 'DEBUG',
            'handlers': ['server_file'],
            'propagate': False,
        },
        'pyx3270.offline': {
            'level': 'DEBUG',
            'handlers': ['offline_file'],
            'propagate': False,
        },
    },
}



import errno
import logging.config
import math
import os
import re
import socket
import subprocess
import sys
from contextlib import closing
from functools import cache
from importlib.resources import files
from logging import getLogger
from time import sleep, time
from typing import Literal

from pyx3270.exceptions import (
    CommandError,
    FieldTruncateError,
    KeyboardStateError,
    NotConnectedException,
    TerminatedError,
)
from pyx3270.iemulator import (
    AbstractCommand,
    AbstractEmulator,
    AbstractEmulatorCmd,
    AbstractExecutableApp,
)
from pyx3270.logging_config import LOGGING_CONFIG
from pyx3270.x3270_commands import x3270_command

logger = getLogger(__name__)


def get_binary_path(*parts):
    # Se estiver rodando como executável PyInstaller
    if hasattr(sys, '_MEIPASS'):
        return os.path.join(sys._MEIPASS, 'pyx3270', 'bin', *parts)

    # Se estiver rodando como pacote instalado normalmente
    try:
        bin_dir = files('pyx3270').joinpath('bin')
        return str(bin_dir.joinpath(*parts))
    except Exception as e:
        raise FileNotFoundError(f'Não foi possível localizar o binário: {e}')


BINARY_FOLDER = get_binary_path()
MODEL_TYPE = Literal['2', '3', '4', '5']
MODEL_DIMENSIONS = {
    '2': {
        'rows': 24,
        'columns': 80,
    },
    '3': {
        'rows': 32,
        'columns': 80,
    },
    '4': {
        'rows': 43,
        'columns': 80,
    },
    '5': {
        'rows': 27,
        'columns': 132,
    },
}


class ExecutableApp(AbstractExecutableApp):
    args = list()

    def __init__(self, shell: bool = False, model: MODEL_TYPE = '2') -> None:
        logger.debug(f'Inicializando ExecutableApp ({shell=}, {model=})')
        self.shell = shell
        self.subprocess = None
        self.args = self._get_executable_app_args(model)
        self._spawn_app()

    def _spawn_app(self, args=None) -> None:
        logger.debug('Iniciando processo do aplicativo')
        kwargs = {
            'shell': self.shell,
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.PIPE,
        }

        if args:
            self.args = args

        if os.name == 'nt':
            logger.debug(
                'Detectado sistema Windows, configurando flags específicos'
            )
            kwargs['creationflags'] = subprocess.CREATE_NO_WINDOW
        else:
            logger.debug(
                'Detectado sistema não-Windows, configurando nova sessão'
            )
            kwargs['start_new_session'] = True

        try:
            logger.debug(f'Executando comando: {self.args}')
            self.subprocess = subprocess.Popen(self.args, **kwargs)
            logger.debug(f'Processo iniciado com PID: {self.subprocess.pid}')
        except Exception:
            logger.error('Erro ao iniciar processo')
            raise

    def _get_executable_app_args(self, model: MODEL_TYPE) -> list:
        logger.debug(f'Obtendo argumentos para modelo: {model}')
        args = self.__class__.args + [
            '-xrm',
            f'*model:{model}',
            '-utf8',
        ]
        logger.debug(f'Argumentos completos: {args}')
        return args

    def connect(*args) -> bool:
        logger.debug(f'Método connect chamado com args: {args}')
        return False

    def close(self):
        logger.info('Fechando aplicativo')
        if self.subprocess and self.subprocess.poll() is None:
            logger.debug('Terminando processo em execução')
            self.subprocess.terminate()
        return_code = self.subprocess.returncode or self.subprocess.poll()
        return_code = return_code if return_code is not None else 0
        logger.info(f'Aplicativo fechado com código de retorno: {return_code}')
        return return_code

    def write(self, data: str):
        logger.debug(f'Escrevendo dados para o processo: {data}')
        try:
            self.subprocess.stdin.write(data)
            self.subprocess.stdin.flush()
            logger.debug('Dados escritos com sucesso')
        except Exception:
            logger.error('Erro ao escrever dados')
            raise

    def readline(self, timeout=5) -> bytes:
        try:
            logger.debug('Aguardando dados no buffer do processo')
            line = self.subprocess.stdout.readline()
            logger.debug(f'Linha lida: {line}')
            return line
        except Exception:
            logger.error('Erro ao ler linha')
            raise


class Command(AbstractCommand):
    def __init__(self, app: ExecutableApp, cmdstr: bytes | str) -> None:
        logger.debug(f'Inicializando Command com comando: {cmdstr}')
        if isinstance(cmdstr, str):
            cmdstr = bytes(cmdstr, 'utf-8', errors='replace')
        self.app = app
        self.cmdstr = cmdstr
        self.status_line = None
        self.data = []

    def execute(self) -> bool:
        logger.debug(f'Executando comando: {self.cmdstr}')
        try:
            self.app.write(self.cmdstr + b'\n')

            while True:
                line = self.app.readline()
                if not line.startswith('data:'.encode('utf-8')):
                    self.status_line = line.rstrip()
                    logger.debug(f'Status line: {self.status_line}')
                    result = self.app.readline().rstrip()
                    logger.debug(f'Resultado: {result}')
                    return self.handle_result(result.decode('utf-8'))

                logger.debug(f'Dados recebidos: {line}')
                self.data.append(line[6:].rstrip('\n\r'.encode('utf-8')))

        except Exception:
            logger.error(f'Erro durante execução do comando: {self.cmdstr}')
            raise

    def handle_result(self, result: str) -> bool:
        logger.debug(f'Processando resultado: {result}')
        count = 0
        max_loop = 5
        while count < max_loop:
            try:
                if not result and self.cmdstr == b'Quit':
                    logger.info('Comando Quit executado com sucesso')
                    return True
                elif result.lower() == 'ok':
                    logger.debug('Comando executado com sucesso (OK)')
                    return True
                else:
                    error_msg = f'"erro" esperado, mas recebido: {result}.'
                    logger.warning(error_msg)
                    raise ValueError(error_msg)
            except ValueError:
                logger.warning(
                    f'Tentativa {count + 1}/{max_loop} falhou, aguardando 1s'
                )
                sleep(0.35)
                count += 1

        msg = b'[sem mensagem de erro]'
        if self.data:
            msg = ''.encode('utf-8').join(self.data).rstrip()
        error_msg = msg.decode('utf-8')

        if (
            'keyboard locked' in error_msg.lower()
            or 'canceled' in error_msg.lower()
        ):
            logger.error(f'Teclado travado detectado: {error_msg}')
            raise KeyboardStateError(error_msg)

        logger.error(f'Comando falhou: {error_msg}')
        raise CommandError(error_msg)


class Status:
    def __init__(self, status_line: str) -> None:
        logger.debug(f'Inicializando Status com linha: {status_line}')
        if not status_line:
            status_line = (' ' * 12).encode('utf-8')
            logger.debug('Status line vazia, usando padrão')
        self.status_line = status_line
        parts = status_line.split(' '.encode('utf-8'))

        try:
            self.keyboard = parts[0] or None
            self.screen_format = parts[1] or None
            self.field_protection = parts[2] or None
            self.connection_state = parts[3] or None
            self.emulator_mode = parts[4] or None
            self.model_number = parts[5] or None
            self.row_number = parts[6] or None
            self.col_number = parts[7] or None
            self.cursor_row = parts[8] or None
            self.cursor_col = parts[9] or None
            self.window_id = parts[10] or None
            self.exec_time = parts[11] or None
            logger.debug(
                f'Status: {self.connection_state=}, {self.emulator_mode=}'
            )
        except IndexError:
            logger.error('Status não tem items suficientes.')

    def __str__(self) -> str:
        return f'Status: {self.status_line}'


class Wc3270App(ExecutableApp):
    args = ['-xrm', '"wc3270.unlockDelay: False"']

    def __init__(self, model: MODEL_TYPE) -> None:
        logger.info(f'Inicializando Wc3270App com modelo: {model}')
        self.args = self._get_executable_app_args(model)
        self.script_port = Wc3270App._get_free_port()
        logger.debug(f'Porta de script alocada: {self.script_port}')
        super().__init__(shell=True, model=model)

    @staticmethod
    def _get_free_port() -> str:
        logger.debug('Obtendo porta livre para comunicação')
        try:
            with closing(
                socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            ) as s:
                s.bind(('127.0.0.1', 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                port = s.getsockname()[1]
                logger.debug(f'Porta livre obtida: {port}')
                return port
        except Exception:
            logger.error('Erro ao obter porta livre.')
            raise

    @cache
    def _make_socket(self) -> None:
        logger.info(f'Criando socket para porta {self.script_port}')
        self.socket = sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        count = 0
        max_loop = 5
        while count < max_loop:
            try:
                logger.debug(
                    f'Tentativa {count + 1}/{max_loop} port:{self.script_port}'
                )
                sock.connect(('localhost', self.script_port))
                logger.info('Conexão de socket estabelecida com sucesso')
                break
            except socket.error as e:
                if e.errno != errno.ECONNREFUSED:
                    logger.error('Erro de conexão não recuperável.')
                    raise NotConnectedException
                logger.warning(
                    f'Conexão recusada, tentando novamente em 1s '
                    f'({count + 1}/{max_loop})'
                )
                sleep(1)
                count += 1
                if count >= max_loop:
                    logger.error(
                        f'Falha ao conectar após {max_loop} tentativas'
                    )

        self.socket_fh = sock.makefile(mode='rwb')
        logger.debug('File handle do socket criado')

    def connect(self, host: str) -> bool:
        logger.info(f'Conectando ao host: {host}')
        self.args = [
            'start',
            '/wait',
            '""',
            f'"{get_binary_path("windows", "wc3270")}"',
        ] + self.args
        self.args.extend(['-scriptport', str(self.script_port), host])
        logger.debug(f'Argumentos completos: {self.args}')

        try:
            self._spawn_app(' '.join(self.args))
            self._make_socket()
            return True
        except Exception:
            logger.error(f'Erro ao conectar ao host: {host}')
            raise

    def close(self) -> None:
        logger.info('Fechando conexão de socket')
        try:
            self.socket.close()
            logger.debug('Socket fechado com sucesso')
        except Exception:
            logger.error('Erro ao fechar socket.')

    def write(self, data: str) -> None:
        logger.debug(f'Escrevendo dados para socket: {data}')
        if self.socket_fh is None:
            logger.error('Tentativa de escrita em socket não inicializado')
            raise NotConnectedException
        try:
            self.socket_fh.write(data)
            self.socket_fh.flush()
            logger.debug('Dados escritos com sucesso')
        except OSError:
            logger.error('Erro de E/S ao escrever no socket')
            raise NotConnectedException

    def readline(self) -> bytes:
        logger.debug('Lendo linha do socket')
        if self.socket_fh is None:
            logger.error('Tentativa de leitura de socket não inicializado')
            raise NotConnectedException
        try:
            line = self.socket_fh.readline()
            logger.debug(f'Linha lida: {line}')
            return line
        except Exception:
            logger.error('Erro ao ler do socket')
            raise NotConnectedException


class Ws3270App(ExecutableApp):
    args = [
        get_binary_path('windows', 'ws3270'),
        '-xrm',
        'ws3270.unlockDelay:False',
    ]

    def __init__(self, model: MODEL_TYPE) -> None:
        logger.info(f'Inicializando Ws3270App com modelo: {model}')
        super().__init__(shell=False, model=model)


class X3270App(ExecutableApp):
    args = [
        get_binary_path('linux', 'x3270'),
        '-xrm',
        'x3270.unlockDelay:False',
        '-script',
    ]

    def __init__(self, model: MODEL_TYPE) -> None:
        logger.info(f'Inicializando X3270App com modelo: {model}')
        super().__init__(shell=False, model=model)


class S3270App(ExecutableApp):
    args = [
        get_binary_path('linux', 's3270'),
        '-xrm',
        's3270.unlockDelay:False',
    ]

    def __init__(self, model: MODEL_TYPE) -> None:
        logger.info(f'Inicializando S3270App com modelo: {model}')
        super().__init__(shell=True, model=model)


class X3270Cmd(AbstractEmulatorCmd):
    def __init__(self, time_unlock: int = 60) -> None:
        logger.info(f'Inicializando X3270Cmd com time_unlock: {time_unlock}')
        self.time_unlock = time_unlock

    def __getattr__(self, name):
        def x3270_builtin_func(*args, **kwargs):
            return x3270_command(self, name, *args, **kwargs)

        return x3270_builtin_func

    def clear_screen(self) -> None:
        logger.info('Limpando tela')
        count = 0
        max_loop = 6
        while count < max_loop:
            logger.debug(f'Tentativa {count + 1}/{max_loop} de limpar tela')
            self.clear()
            self.wait(self.time_unlock, 'unlock')
            if not self.get_full_screen(header=True).strip():
                logger.info('Tela limpa com sucesso')
                break
            logger.warning(
                f'Tela não foi limpa completamente na tentativa {count + 1}'
            )
            count += 1
        if count >= max_loop:
            logger.warning(
                f'Não foi possível limpar a tela completamente'
                f' após {max_loop} tentativas'
            )

    def wait_for_field(self, timeout: int = 30) -> None:
        logger.debug(f'Aguardando campo de entrada (timeout={timeout}s)')
        try:
            self.wait(timeout, 'InputField')
            logger.debug('Campo de entrada carregado.')
        except CommandError:
            logger.warning(
                f'Timeout atingido: {timeout}s.campo entrada não encontrado.'
            )

    def wait_string_found(
        self,
        ypos: int,
        xpos: int,
        string: str,
        equal: bool = True,
        timeout: int = 5,
    ) -> bool:
        logger.debug(
            f'Aguardando {string=} na posição '
            f'({ypos},{xpos}), {equal=}, {timeout=}s'
        )

        end_time = time() + timeout
        result = None

        while time() < end_time:
            try:
                found = self.get_string(ypos, xpos, len(string))
                logger.debug(f"String encontrada: '{found}'")
                if equal:
                    result = found == string
                else:
                    result = found != string
                logger.debug(f'Resultado da comparação: {result}')
                if result:
                    break
            except Exception:
                logger.debug('Erro ao buscar string, tentando novamente')
                continue

        logger.warning(
            f'Timeout atingido após {timeout}s, resultado final: {result}'
        )
        return result

    def string_found(self, ypos: int, xpos: int, string: str) -> bool:
        logger.debug(
            f'Verificando se {string=} existe na posição ({ypos},{xpos})'
        )
        try:
            found = self.get_string(ypos, xpos, len(string))
            result = found == string
            logger.debug(f"Resultado: {result} (encontrado: '{found}')")
            return result
        except Exception:
            logger.error(
                f'Erro ao verificar {string=}',
            )
            return False

    def delete_field(self) -> None:
        logger.debug('Deletando campo atual')
        self.deletefield()
        logger.debug('Campo deletado')

    def move_to(self, ypos: int, xpos: int) -> None:
        logger.debug(f'Movendo cursor para posição ({ypos},{xpos})')
        self.movecursor1(ypos, xpos)
        logger.debug('Cursor movido')

    def send_pf(self, value: int) -> None:
        logger.info(f'Enviando tecla PF{value}')
        self.pf(value)
        logger.debug(f'PF{value} enviado e tela desbloqueada')

    def send_string(
        self,
        tosend: str,
        ypos: int | None = None,
        xpos: int | None = None,
        password: bool = False,
    ) -> None:
        if not tosend:
            logger.warning('tosend não é string, send_string não executado.')
            return
        # Remove caracteres especiais
        original = tosend
        tosend = re.sub(r"[()\"']", '', tosend)

        tosend_str = 'password' if password else tosend

        if original != tosend:
            logger.debug(
                f'String modificada para {tosend_str} '
                f'(removidos caracteres especiais)'
            )

        if xpos is not None and ypos is not None:
            logger.info(
                f"Enviando string '{tosend_str}' para posição {ypos=} {xpos=}"
            )
            self.move_to(ypos, xpos)
        else:
            logger.info(f"Enviando string '{tosend_str}' na posição atual.")

        self.string(f'"{tosend}"')
        self.wait(self.time_unlock, 'unlock')
        logger.debug("String '{tosend_str}' enviada para o emulador.")

    def send_enter(self) -> None:
        logger.info('Enviando tecla ENTER')
        self.enter()
        self.wait(self.time_unlock, 'unlock')
        logger.debug('ENTER enviado e tela desbloqueada')

    def send_home(self) -> None:
        logger.info('Enviando tecla HOME')
        self.home()
        self.wait(self.time_unlock, 'unlock')
        logger.debug('HOME enviado e tela desbloqueada')

    def get_string(self, ypos: int, xpos: int, length: int) -> str:
        logger.debug(
            f'Obtendo string na posição ({ypos},{xpos})'
            f' com comprimento {length}'
        )
        try:
            self.check_limits(ypos, xpos)
            if (xpos + length) > (self.model_dimensions['columns'] + 1):
                logger.error(
                    f'Comprimento excede limite da tela: {xpos}+{length}'
                    f' > {self.model_dimensions["columns"] + 1}'
                )
                raise FieldTruncateError

            xpos -= 1
            ypos -= 1
            result = self.ascii(ypos, xpos, length)
            logger.debug(f"String obtida: '{result}'")
            return result
        except Exception:
            logger.error(
                'Erro ao obter string',
            )
            raise

    def get_string_area(
        self, yposi: int, xposi: int, ypose: int, xpose: int
    ) -> str:
        logger.debug(
            f'Obtendo área de texto de ({yposi},{xposi}) até ({ypose},{xpose})'
        )
        try:
            self.check_limits(yposi, xposi)
            self.check_limits(ypose, xpose)
            yposi -= 1
            xposi -= 1
            ypose -= yposi
            xpose -= xposi
            result = self.ascii(yposi, xposi, ypose, xpose)
            logger.debug(f'Área obtida com {len(result)} caracteres')
            return result
        except Exception:
            logger.error('Erro ao obter área de texto')
            raise

    def get_full_screen(self, header: bool = True) -> str:
        logger.debug(
            f'Obtendo conteúdo completo da tela (com header: {header})'
        )
        try:
            text = self.ascii()
            if not header:
                start = self.model_dimensions['columns']
                text = text[start:]
                logger.debug('Header removido do conteúdo')
            logger.debug(f'Conteúdo obtido com {len(text)} caracteres')
            return text
        except Exception:
            logger.error(
                'Erro ao obter conteúdo da tela.',
            )
            raise

    def save_screen(self, file_path: str, file_name: str):
        logger.info(f'Salvando tela em {file_path}\\{file_name}.html')
        try:
            if not os.path.exists(file_path):
                logger.debug(f'Criando diretório: {file_path}')
                os.makedirs(file_path)
            self.printtext('html', 'file', f'{file_path}\\{file_name}.html')
            logger.info('Tela salva com sucesso')
        except Exception:
            logger.error(
                'Erro ao salvar tela.',
            )
            raise

    def check_limits(self, ypos, xpos):
        logger.debug(f'Verificando limites para posição ({ypos},{xpos})')
        if ypos > self.model_dimensions['rows']:
            error_msg = (
                f'Você excedeu o limite do eixo y da tela do mainframe: '
                f'{ypos} > {self.model_dimensions["rows"]}'
            )
            logger.error(error_msg)
            raise FieldTruncateError(error_msg)
        if xpos > self.model_dimensions['columns']:
            error_msg = (
                f'Você excedeu o limite do eixo x da tela do mainframe: '
                f'{xpos} > {self.model_dimensions["columns"]}'
            )
            logger.error(error_msg)
            raise FieldTruncateError(error_msg)
        logger.debug('Posição dentro dos limites')

    def search_string(self, string: str, ignore_case: bool = False) -> bool:
        logger.info(f"Buscando texto '{string}' na tela ({ignore_case=})")
        try:
            for ypos in range(1, self.model_dimensions['rows'] + 1):
                line = self.get_string(
                    ypos, 1, self.model_dimensions['columns']
                )
                if ignore_case:
                    string_comp = string.lower()
                    line_comp = line.lower()
                    logger.debug(
                        f'Comparando (ignorando case) na linha {ypos}'
                    )
                else:
                    string_comp = string
                    line_comp = line
                    logger.debug(
                        f'Comparando (case sensitive) na linha {ypos}'
                    )

                if string_comp in line_comp:
                    logger.info(f'Texto encontrada na linha {ypos}')
                    return True

            logger.info('Texto não encontrada em nenhuma linha')
            return False
        except Exception:
            logger.error('Erro durante busca de texto')
            return False

    def get_string_positions(
        self, string: str, ignore_case=False
    ) -> list[tuple[int]]:
        logger.info(f"Buscando posições da texto '{string}' ({ignore_case=})")
        try:
            screen_content = self.get_full_screen(header=True)
            flags = 0 if not ignore_case else re.IGNORECASE
            indices_object = re.finditer(
                re.escape(string), screen_content, flags
            )
            indices = [index.start() for index in indices_object]
            logger.debug(f'Encontradas {len(indices)} ocorrências')

            positions = [
                self._get_ypos_and_xpos_from_index(index + 1)
                for index in indices
            ]
            logger.info(f'Posições encontradas: {positions}')
            return positions
        except Exception:
            logger.error('Erro ao buscar posições')
            return []

    def _get_ypos_and_xpos_from_index(self, index):
        logger.debug(f'Convertendo índice {index} para coordenadas (y,x)')
        ypos = math.ceil(index / self.model_dimensions['columns'])
        remainder = index % self.model_dimensions['columns']
        if remainder == 0:
            xpos = self.model_dimensions['columns']
        else:
            xpos = remainder
        logger.debug(f'Índice {index} convertido para ({ypos},{xpos})')
        return (ypos, xpos)


class X3270(AbstractEmulator, X3270Cmd):
    def __init__(
        self,
        visible: bool = False,
        model: MODEL_TYPE = '2',
        save_log_file: bool = False,
        time_unlock: int = 60,
    ) -> None:
        if save_log_file:
            logging.config.dictConfig(LOGGING_CONFIG)
        X3270Cmd.__init__(self, time_unlock=time_unlock)
        logger.info(f'Inicializando X3270 (visible={visible}, model={model})')
        self.model = model
        self.model_dimensions = MODEL_DIMENSIONS[model]
        self.visible = visible
        self.app: ExecutableApp = self._create_app()
        self.is_terminated = False
        self.host = None
        self.port = None
        self.tls = None
        self.mode_3270 = None
        self.last_command_time = None
        logger.debug('X3270 inicializado')

    def _create_app(self) -> None:
        logger.info('Criando aplicativo emulador')
        try:
            if os.name == 'nt':  # windows
                if self.visible:
                    logger.debug('Criando Wc3270App (Windows, visível)')
                    return Wc3270App(self.model)
                logger.debug('Criando Ws3270App (Windows, não visível)')
                return Ws3270App(self.model)

            if self.visible:  # linux
                logger.debug('Criando X3270App (Linux, visível)')
                return X3270App(self.model)
            logger.debug('Criando S3270App (Linux, não visível)')
            return S3270App(self.model)

        except Exception:
            logger.error('Erro ao criar aplicativo.')
            raise

    def _exec_command(self, cmdstr: str) -> Command:
        logger.debug(f'Executando comando: {cmdstr}')
        if self.is_terminated:
            error_msg = 'Tentativa de executar comando em emulador terminado'
            logger.error(error_msg)
            raise TerminatedError
        max_loop = 3
        for exec in range(max_loop):
            try:
                cmd = Command(self.app, cmdstr)
                cmd.execute()
                self.status = Status(cmd.status_line)
                self.last_command_time = time()
                logger.debug(f'Comando executado, status: {self.status}')
                return cmd
            except NotConnectedException:
                logger.error('Emulador não conectado.')
                raise NotConnectedException
            except KeyboardStateError:
                sleep(1)
                logger.warning(
                    f'Nova tentativa de exec command:'
                    f'{cmdstr} {exec + 1}/{max_loop}'
                )
                self.reset()
                self.wait(self.time_unlock, 'unlock')
                self.tab()
        logger.error(
            f'Erro ao executar {cmdstr} total de tentativas: {max_loop}'
        )
        raise CommandError

    def terminate(self) -> None:
        logger.info('Terminando emulador')
        if not self.is_terminated:
            try:
                logger.debug('Enviando comando quit')
                self.quit()
            except BrokenPipeError:
                logger.warning('BrokenPipeError ao enviar quit, ignorando')
                self.ignore()
            except socket.error as ex:
                if ex.errno != errno.ECONNRESET:
                    logger.error('Erro de socket ao terminar')
                    raise ConnectionError
                logger.warning('Erro de conexão resetada')

        logger.debug('Fechando aplicativo')
        self.app.close()
        self.is_terminated = True
        logger.info('Emulador terminado com sucesso')

    def is_connected(self) -> bool:
        logger.debug('Verificando estado de conexão')
        try:
            # Verifica tempo desde o último comando
            elapsed_max = 600
            if self.last_command_time:
                elapsed = time() - self.last_command_time
                if elapsed > elapsed_max:
                    logger.warning(
                        f'Tempo de inatividade excedido: '
                        f'{elapsed:.2f} segundos'
                    )
                    return False

            self.query('ConnectionState')
            is_connected = self.status.connection_state.startswith(b'C(')
            logger.info(f'Estado de conexão: {is_connected}')
            return is_connected
        except Exception:
            logger.error('Erro ao verificar conexão')
            return False

    def connect_host(
        self,
        host: str,
        port: int | str,
        tls: bool = True,
        mode_3270: bool = True,
    ) -> None:
        logger.info(f'Conectando ao host: {host}:{port} (tls={tls})')
        self.host = host
        self.port = port
        self.tls = tls
        self.mode_3270 = mode_3270
        tls_prefix = 'L:Y:' if tls else ''
        strint_conn = f'{tls_prefix}{host}:{port}'
        logger.debug(f'String de conexão: {strint_conn}')

        try:
            if self.app:
                if not self.app.connect(strint_conn):
                    logger.debug(
                        'Método connect do app retornou False, '
                        + 'tentando método connect direto'
                    )
                    self.connect(strint_conn)
                if mode_3270:
                    logger.debug('Aguardando modo 3270')
                    self.wait(5, '3270mode')
                logger.info('Conexão estabelecida com sucesso')
        except CommandError:
            logger.warning('CommandError durante conexão')
        except Exception:
            logger.error('Erro ao conectar')
            raise

    def reconnect_host(self) -> 'X3270':
        logger.info('Tentando reconectar ao host')
        try:
            logger.debug('Executando comando reconnect')
            self.reconnect()
            logger.info('Reconexão bem-sucedida')
            return self
        except Exception:
            logger.warning('Erro durante reconexão.')
            logger.debug('Terminando instância atual')
            self.terminate()
        finally:
            logger.info('Criando nova instância para reconexão')
            args = self.host, self.port, self.tls, self.mode_3270
            logger.debug(f'Argumentos para nova instância: {args}')
            new_instance = X3270(self.visible, self.model)
            new_instance.connect_host(*args)
            logger.debug('Nova instância criada com sucesso')
            # Atualiza todos os atributos de self com os do novo objeto
            self.__dict__.update(new_instance.__dict__)

            logger.debug('Atributos de self atualizados com sucesso')
            return self
